/*
 * gamemanager.c
 *
 * Copyright (c) 2001- Kenta Cho. All rights reserved.
 */
#include "galib.h"
#include "bslib.h"
#include "meteo.h"
#include <keyiolib.h>

/**
 * Handle game status.
 *
 * @author cs8k-cyu@asahi-net.or.jp
 * @version 0.21
 */
GameStatus gameStatus;

#define METEO_MAX 32
#define LASER_MAX 16
#define FRAG_MAX 96
#define LINE_MAX 16

#define NOT_EXIST -9999

static Meteo meteoInst[METEO_MAX];
static Meteo *meteo[METEO_MAX];
static Laser laser[LASER_MAX];
static Frag frag[FRAG_MAX];
static int score, mtSc, left;

static void initGame() {
  int i;
  for ( i=0 ; i<METEO_MAX ; i++ ) {
    meteoInst[i].z = NOT_EXIST;
    meteo[i] = &(meteoInst[i]);
  }
  for ( i=0 ; i<LASER_MAX ; i++ ) {
    laser[i].idx = NOT_EXIST;
  }
  for ( i=0 ; i<FRAG_MAX ; i++ ) {
    frag[i].cnt = NOT_EXIST;
  }
  mtSc = 0;
}

static void startTitle() {
  gameStatus.status = TITLE;
  drawTitleFirst();
  initGame();
}

void drawLeft() {
  drawNum(260, left);
}

void drawScore() {
  drawNum(0, score);
}

static void startGame() {
  gameStatus.status = GAME_IN_PROGRESS;
  score = 0;
  left = 3;
  drawView();
  drawLeft();
  drawScore();
  initGame();
}

static void restartGame() {
  gameStatus.status = GAME_IN_PROGRESS;
  initGame();
  mtSc = score;
}

static void startGameover() {
  gameStatus.status = GAME_OVER;
  gameStatus.cnt = 0;
  drawGameover();
  initGame();
}

void init() {
  initSound();
  makeModels();
  startTitle();
}

/*
 * Objects(meteo, laser, ...) moves.
 */
static int meteoIdx = 0;
extern GaModel meteoModel[METEO_MODEL_NUM];

static void calcMeteoMove(Meteo *mt, int ax, int ay, int az, int speed) {
  mt->mx = (ax-mt->x) / speed;
  mt->my = (ay-mt->y) / speed;
  mt->mz = (az-mt->z) / speed;
}

static void calcMeteoAimMove(Meteo *mt, int speed) {
  mt->mx += ((((mt->ax-mt->x) / speed) - mt->mx)>>4);
  mt->my += ((((mt->ay-mt->y) / speed) - mt->my)>>4);
  mt->mz += ((((mt->az-mt->z) / speed) - mt->mz)>>4);
}

static void addMeteo(int speed, int type) {
  int i;
  DVAL ax, ay, az;
  Meteo* mt;
  meteoIdx++; meteoIdx &= (METEO_MAX-1);
  if ( meteo[meteoIdx]->z != NOT_EXIST ) return;
  mt = meteo[meteoIdx];
  if ( speed <= 0 ) speed = 1;
  mt->type = type;
  switch ( type ) {
  case 0:
    mt->x = DV(-256 + randN(512));
    mt->y = DV(320 + randN(32));
    mt->z = DV(4096 + randN(256));
    ax = DV(-80 + randN(160));
    ay = DV(-256);
    az = DV(512);
    break;
  case 1:
    mt->x = DV(-512 + randN(1024));
    mt->y = DV(-256 - randN(256));
    mt->z = DV(8192 + randN(1024));
    ax = DV(-256 + randN(512));
    ay = DV(120+randN(360));
    az = DV(5120);
    speed -= (speed>>2);
    mt->cnt = mt->speed = speed;
    break;
  }
  calcMeteoMove(mt, ax, ay, az, speed);
  mt->d1 = randN(360); mt->d2 = randN(360);
  mt->md1 = -8 + randN(16);
  mt->md2 = -8 + randN(16);
  mt->model = &(meteoModel[randN(METEO_MODEL_NUM)]);
  mt->prvLockedKey = NOT_EXIST;
  mt->sy = 80;
  for ( i=0 ; i<TAIL_NUM ; i++ ) {
    mt->tailDeg[i] = randN(360);
    mt->tailDegMv[i] = randN(15)-7;
  }
}

static int addCnt = 0;

static void addMeteos() {
  int rank, adSpd, cntBs, type;
  addCnt--;
  if ( addCnt > 0 ) return;
  rank = (mtSc%100)*((int)(mtSc/100)+3)+1;
  if ( rank > 2400 ) rank = 2400;
  mtSc++;
  adSpd = 256000/(rank+1000); adSpd += randN(adSpd>>1);
  if ( randN(5000/rank) == 0 ) {
    adSpd >>= 2;
  } else if ( randN(1400/rank) == 0 ) {
    adSpd >>= 1;
  }
  if ( randN((500/rank)+2)==0 ) {
    type = 1;
  } else {
    type = 0;
  }
  addMeteo(adSpd, type);
  cntBs = 8-(mtSc/100); if ( cntBs < 2 ) cntBs = 2;
  addCnt = (20-(mtSc%100)/5+cntBs) + randN(cntBs);
}

static int fragIdx = 0;

static void addFrag(DVAL x, DVAL y, DVAL z, 
		    DVAL mx, DVAL my, DVAL mz, int type) {
  Frag *fr;
  int i;
  fragIdx++; fragIdx &= (FRAG_MAX-1);
  fr = &(frag[fragIdx]);
  fr->x = x; fr->y = y; fr->z = z;
  fr->mx = mx; fr->my = my; fr->mz = mz;
  fr->d1 = randN(360); fr->d2 = randN(360);
  fr->md1 = randN(50)-25; fr->md2 = randN(50)-25;
  fr->type = type;
  switch ( type ) {
  case 0:
    fr->cnt = 16+randN(64);
    break;
  case 1:
  case 2:
    fr->cnt = 20+randN(10);
    break;
  }
}

#define LASER_SHOT_POS 48
#define LASER_SHOT_MV 20

static int laserIdx = 0;
static int laserCnt = 0;

static void addLaser(Meteo* aim) {
  Laser* ls;
  int i;
  for ( i=0 ; i<LASER_MAX ; i++ ) {
    laserIdx++; laserIdx &= (LASER_MAX-1);
    if ( laser[laserIdx].idx == NOT_EXIST ) break;
  }
  if ( i >= LASER_MAX ) return;
  ls = &(laser[laserIdx]);
  ls->mz = DV(120);
  ls->z[0] = DV(32);
  switch ( laserCnt ) {
  case 0:
    ls->x[0] = DV(-LASER_SHOT_POS); ls->y[0] = DV(-LASER_SHOT_POS); 
    ls->mx = DV(LASER_SHOT_MV); ls->my = DV(LASER_SHOT_MV); 
    break;
  case 1:
    ls->x[0] = DV(LASER_SHOT_POS); ls->y[0] = DV(-LASER_SHOT_POS); 
    ls->mx = DV(-LASER_SHOT_MV); ls->my = DV(LASER_SHOT_MV);
    break;
  case 2:
    ls->x[0] = DV(-LASER_SHOT_POS); ls->y[0] = DV(LASER_SHOT_POS);
    ls->mx = DV(LASER_SHOT_MV); ls->my = DV(-LASER_SHOT_MV);
    break;
  case 3:
    ls->x[0] = DV(LASER_SHOT_POS); ls->y[0] = DV(LASER_SHOT_POS);
    ls->mx = DV(-LASER_SHOT_MV); ls->my = DV(-LASER_SHOT_MV);
    break;
  }
  ls->idx = 0; ls->showAll = FALSE;
  ls->aim = aim;
  ls->lockCnt = 10;
  laserCnt++; laserCnt &= 3;
}

#define KEY_CODE_MAX 4

static DKEYCODE inputKey[KEY_CODE_MAX+1], prvInputKey[KEY_CODE_MAX+1];
static int prvKeyNum = 0;
static DKEYCODE pushedKey[KEY_CODE_MAX+1];
static int pushedKeyNum;

static void checkKey() {
  int keyNum, i, j;
  BOOLEAN stillPressed;
  keyNum = GetPushingKey(inputKey, KEY_CODE_MAX);
  pushedKeyNum = 0;

  for ( i=0 ; i<keyNum ; i++ ) {
    stillPressed = FALSE;
    for ( j=0 ; j<prvKeyNum ; j++ ) {
      if ( inputKey[i] == prvInputKey[j] ) {
	stillPressed = TRUE;
	break;
      }
    }
    if ( !stillPressed ) {
      pushedKey[pushedKeyNum] = inputKey[i];
      pushedKeyNum++;
    }
  }
  for ( i=0 ; i<keyNum ; i++ ) {
    prvInputKey[i] = inputKey[i];
  }
  prvKeyNum = keyNum;
}

static int keyPos[] = {
  -1,-1,-1,-1,-1,-1,-1,-1,  19,-1,-1,-1,-1,35,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,  -1,-1,-1,-1,-1,-1,-1,-1,
  34,-1,-1,-1,-1,-1,-1,-1,  -1,-1,-1,-1,28,29,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,  -1,-1,-1,-1,-1,-1,-1,-1,
  31,10,25,23,12, 2,13,14,  15, 7,16,17,18,27,26, 8,
   9, 0, 3,11, 4, 6,24, 1,  22, 5,21,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,33,-1,-1,  -1,-1,-1,32,-1,-1,20,-1,
  -1,-1,-1,-1,-1,-1,30,-1,  -1,-1,-1,-1,36,-1,-1,-1,
};

#define LASER_SIGHT_MAX 4
#define LASER_SIGHT_WIDTH 50
static LaserSight laserSight[LASER_SIGHT_MAX];
static int lsNum = 0;

static void shotLaser() {
  int i, j;
  int key;
  int lx, ly;
  int mtIdx;
  Meteo *mt, *fstMt = NULL;
  BOOLEAN locked = FALSE;
  for ( i=0 ; i<pushedKeyNum ; i++ ) {
    key = keyPos[pushedKey[i]];
    if ( key == -1 ) continue;
    if ( key < 30 ) {
      lx = (key%10)*(160-LASER_SIGHT_WIDTH)/9; 
      ly = ((int)(key/10))*(160-LASER_SIGHT_WIDTH)/3;
    } else {
      lx = (key-30)*(160-LASER_SIGHT_WIDTH)/6; ly = (160-LASER_SIGHT_WIDTH);
    }
    if ( lsNum < LASER_SIGHT_MAX ) {
      laserSight[lsNum].x = lx;
      laserSight[lsNum].y = ly;
      lsNum++;
    }
    mtIdx = meteoIdx-1;
    for ( j=0 ; j<METEO_MAX ; j++ ) {
      mtIdx++; mtIdx &= (METEO_MAX-1);
      if ( meteo[mtIdx]->z == NOT_EXIST ) continue;
      mt = meteo[mtIdx];
      if ( mt->sx>=lx && mt->sx<lx+LASER_SIGHT_WIDTH && 
	   mt->sy>=ly && mt->sy<ly+LASER_SIGHT_WIDTH ) {
	locked = TRUE;
	if ( mt->prvLockedKey != key ) {
	  mt->prvLockedKey = key;
	  addLaser(mt);
	  break;
	}
      }
    }
    if ( locked && j==METEO_MAX ) {
      mtIdx = meteoIdx-1;
      for ( j=0 ; j<METEO_MAX ; j++ ) {
	mtIdx++; mtIdx &= (METEO_MAX-1);
	if ( meteo[mtIdx]->z == NOT_EXIST ) continue;
	mt = meteo[mtIdx];
	if ( mt->prvLockedKey==key && 
	     mt->sx>=lx && mt->sx<lx+LASER_SIGHT_WIDTH && 
	     mt->sy>=ly && mt->sy<ly+LASER_SIGHT_WIDTH ) {
	  if ( fstMt == NULL ) {
	    fstMt = mt;
	  } else {
	    mt->prvLockedKey = -1;
	  }
	}
      }
      if ( fstMt != NULL ) addLaser(fstMt);
    }
  }
}

static void sortMeteos() {
  int i, j;
  int idx, stIdx = meteoIdx;
  Meteo *vmt;
  DVAL vz;
  for ( i=0 ; i<METEO_MAX-1 ; i++ ) {
    idx = stIdx-i-2; idx &= METEO_MAX-1;
    vmt = meteo[idx]; vz = vmt->z;
    for ( j=0 ; j<=i ; j++ ) {
      idx++; idx &= METEO_MAX-1;
      if ( vz > meteo[idx]->z ) {
	meteo[(idx-1)&(METEO_MAX-1)] = meteo[idx];
      } else {
	meteo[(idx-1)&(METEO_MAX-1)] = vmt;
	break;
      }
    }
    if ( j == i+1 ) {
      meteo[idx] = vmt;
    }
  }
}

static void addMeteoFrags(Meteo *mt) {
  int i;
  for ( i=0 ; i<8+randN(8) ; i++ ) {
    addFrag(mt->x, mt->y, mt->z,
	    mt->mx + DV(randN(7)-3),
	    mt->my + DV(randN(7)-3),
	    mt->mz + DV(randN(7)-3), 0);
  }
  mt->z = NOT_EXIST;
}

static void hitted() {
  int i;
  gameStatus.status = HIT;
  gameStatus.cnt = 0;
  left--;
  drawLeft();
  for ( i=0 ; i<METEO_MAX ; i++ ) {
    if ( meteo[i]->z == NOT_EXIST ) continue;
    addMeteoFrags(meteo[i]);
  }
}

static void destroyMeteo(Meteo *mt) {
  score++; if ( score > 9999 ) score = 9999;
  drawScore();
  addMeteoFrags(mt);
}

static void moveMeteos() {
  int i, j;
  Meteo* mt;
  for ( i=0 ; i<METEO_MAX ; i++ ) {
    if ( meteo[i]->z == NOT_EXIST ) continue;
    mt = meteo[i];
    switch ( mt->type ) {
    case 0:
      if ( mt->sy > 160 ) {
	if ( gameStatus.status != GAME_IN_PROGRESS ) {
	  mt->z = NOT_EXIST;
	  continue;
	} else if ( mt->prvLockedKey != NOT_EXIST 
		    || mt->sx<0 || mt->sx>160 ) {
	  destroyMeteo(mt);
	  continue;
	} else {
	  hitted();
	  return;
	}
      }
      break;
    case 1:
      mt->cnt--;
      mt->my -= (mt->my>>9);
      if ( mt->cnt < 0 ) {
	calcMeteoAimMove(mt, mt->speed);
	if ( mt->cnt<-31 || mt->sy<-10 ) {
	  calcMeteoMove(mt, mt->ax, mt->ay, mt->ax, mt->speed);
	  mt->type = 0;
	}
      } else if ( mt->cnt==0 || mt->sy<20 ) {
	if ( mt->sy < 20 ) {
	  if ( mt->x < 0 ) {
	    mt->ax = DV(10 + randN(10));
	  } else {
	    mt->ax = DV(-10 - randN(10));
	  }
	} else {
	  mt->ax = DV(-20 + randN(40));
	}
	mt->ay = DV(-256);
	mt->az = DV(512);
	mt->cnt = 0;
      }
      break;
    }
    mt->x += mt->mx;
    mt->y += mt->my;
    mt->z += mt->mz;
    mt->d1 += mt->md1;
    mt->d2 += mt->md2;
    for ( j=0 ; j<TAIL_NUM ; j++ ) {
      mt->tailDeg[j] += mt->tailDegMv[j];
    }
    if ( randN(4) == 0 ) {
      addFrag(mt->x, mt->y, mt->z, 
	      -mt->mx+DV(randN(5)-2),
	      -mt->my+DV(randN(5)-2),
	      -mt->mz+DV(randN(5)-2), 2);
    }
  }
}

static void moveLasers() {
  int i, j;
  int rt, pLsIdx, lsIdx;
  Laser* ls;
  for ( i=0 ; i<LASER_MAX ; i++ ) {
    if ( laser[i].idx == NOT_EXIST ) continue;
    ls = &(laser[i]);
    pLsIdx = ls->idx;
    ls->idx++; ls->idx &= (LASER_LENGTH-1);
    lsIdx = ls->idx;
    if ( !ls->showAll && lsIdx == 0 ) ls->showAll = TRUE;
    rt = (ls->aim->z - ls->z[pLsIdx]) / ls->mz;
    if ( rt <= 0 ) rt = 1;
    ls->mx += (((ls->aim->x - ls->x[pLsIdx])/rt) - ls->mx)>>2;
    ls->my += (((ls->aim->y - ls->y[pLsIdx])/rt) - ls->my)>>2;
    ls->x[lsIdx] = ls->x[pLsIdx] + ls->mx;
    ls->y[lsIdx] = ls->y[pLsIdx] + ls->my;
    ls->z[lsIdx] = ls->z[pLsIdx] + ls->mz;
    if ( ls->lockCnt > 0 ) ls->lockCnt--;
    if ( ls->aim->z == NOT_EXIST ) {
      for ( j=0 ; j<2+randN(4) ; j++ ) {
	addFrag(ls->x[lsIdx], ls->y[lsIdx], ls->z[lsIdx],
		ls->mx + randN(3)-1,
		ls->my + randN(3)-1,
		ls->mz + randN(3)-1, 1);
      }
      ls->idx = NOT_EXIST;
    } else if ( ls->z[lsIdx] >= ls->aim->z ) {
      for ( j=0 ; j<3+randN(7) ; j++ ) {
	addFrag(ls->x[lsIdx], ls->y[lsIdx], ls->z[lsIdx],
		ls->mx + randN(5)-2,
		ls->my + randN(5)-2,
		ls->mz + randN(5)-2, 1);
      }
      ls->idx = NOT_EXIST;
      destroyMeteo(ls->aim);
    } 
  }
}

static void moveFrags() {
  int i;
  Frag *fr;
  for ( i=0 ; i<FRAG_MAX ; i++ ) {
    if ( frag[i].cnt < 0 ) continue;
    fr = &(frag[i]);
    fr->cnt--;
    fr->x += fr->mx;
    fr->y += fr->my;
    fr->z += fr->mz;
    fr->d1 += fr->md1;
    fr->d2 += fr->md2;
  }
}

void move() {
  checkKey();
  switch ( gameStatus.status ) {
  case GAME_IN_PROGRESS:
    shotLaser();
    addMeteos();
    break;
  case TITLE:
    if ( pushedKeyNum>0 && keyPos[pushedKey[0]] != -1 ) {
      startGame();
      return;
    }
    addMeteos();
    break;
  case HIT:
    if ( gameStatus.cnt > 64 ) {
      if ( left <= 0 ) {
	startGameover();
      } else {
	restartGame();
      }
      return;
    }
    break;
  case GAME_OVER:
    if ( pushedKeyNum > 0 || gameStatus.cnt > 128 ) {
      startTitle();
    }
    return;
  }
  moveFrags();
  moveMeteos();
  moveLasers();
  sortMeteos();
}


/*
 * Draw objects.
 */
static int lsClrCnt = 0;
static DCOL laserColor[] = {
  RGBA_MAKE(40, 40, 255, 0),
  RGBA_MAKE(140, 240, 255, 0),
  RGBA_MAKE(40, 255, 140, 0),
  RGBA_MAKE(140, 255, 240, 0),
};
static DTRI lsIp[] = {
  0,1,2, 0,2,1, 1,2,3, 1,3,2, 
  2,3,4, 2,4,3, 3,4,5, 3,5,4,
  4,5,6, 4,6,5, 5,6,7, 5,7,6,
};

static DCOL lockColor = RGBA_MAKE(255, 255, 0, 0);
static DTRI lockIp[] = {0,1,2};

#define LASER_WIDTH DV(2)

static void drawLasers() {
  int i, j;
  int lgt, lsIdx, vtxIdx;
  Laser* ls;
  GaMatrix mtr;
  GaVertex vtx[LASER_LENGTH*2];
  GaRVertex rvtx[9];
  Meteo *mt;
  DANG lcd;

  GaLightEnable(FALSE);
  GaSetMaterial(laserColor[lsClrCnt]);
  lsClrCnt++; lsClrCnt &= 3;
  GaMatIdentity(&mtr);
  GaSetTransform(&mtr);
  for ( i=0 ; i<LASER_MAX ; i++ ) {
    if ( laser[i].idx == NOT_EXIST ) continue;
    ls = &(laser[i]);
    if ( ls->showAll ) lgt = LASER_LENGTH;
    else lgt = ls->idx+1;
    lsIdx = ls->idx;
    for ( j=0, vtxIdx=0 ; j<lgt ; j++, vtxIdx+=2 ) {
      vtx[vtxIdx].x = ls->x[lsIdx] - LASER_WIDTH; 
      vtx[vtxIdx+1].x = ls->x[lsIdx] + LASER_WIDTH;
      vtx[vtxIdx].y = vtx[vtxIdx+1].y = ls->y[lsIdx];
      vtx[vtxIdx].z = vtx[vtxIdx+1].z = ls->z[lsIdx];
      lsIdx--; lsIdx &= (LASER_LENGTH-1);
    }
    GaDrawPrimitive(vtx, vtxIdx, lsIp, (lgt-1)*12);
    for ( j=0 ; j<3 ; j++ ) {
      mt = ls->aim;
      lcd = D2R((ls->lockCnt<<4)+j*120);
      rvtx[0].x = DV(mt->sx) + GaMul(GaSin(lcd), DV(4+ls->lockCnt));
      rvtx[0].y = DV(mt->sy) + GaMul(GaCos(lcd), DV(4+ls->lockCnt));
      lcd = D2R((ls->lockCnt<<4)+j*120+20);
      rvtx[1].x = DV(mt->sx) + GaMul(GaSin(lcd), DV(10+ls->lockCnt));
      rvtx[1].y = DV(mt->sy) + GaMul(GaCos(lcd), DV(10+ls->lockCnt));
      lcd = D2R((ls->lockCnt<<4)+j*120-20);
      rvtx[2].x = DV(mt->sx) + GaMul(GaSin(lcd), DV(10+ls->lockCnt));
      rvtx[2].y = DV(mt->sy) + GaMul(GaCos(lcd), DV(10+ls->lockCnt));
      rvtx[0].diffuse = rvtx[1].diffuse = rvtx[2].diffuse = lockColor;
      GaDrawPrimitiveRV(rvtx, 3, lockIp, 3);
    }
  }
}

static DTRI laserIp[] = {0,1,2, 3,4,5, 6,7,8, 9,10,11};

static void drawLaserSights() {
  int i, j, lsx, lsy;
  GaRVertex rvtx[12];
  for ( i=0 ; i<lsNum ; i++ ) {
    lsx = laserSight[i].x; lsy = laserSight[i].y;
    rvtx[0].x = DV(lsx);
    rvtx[0].y = DV(lsy);
    rvtx[1].x = DV(lsx+8);
    rvtx[1].y = DV(lsy);
    rvtx[2].x = DV(lsx);
    rvtx[2].y = DV(lsy+8);
    rvtx[3].x = DV(lsx+LASER_SIGHT_WIDTH-1);
    rvtx[3].y = DV(lsy);
    rvtx[4].x = DV(lsx+LASER_SIGHT_WIDTH-1);
    rvtx[4].y = DV(lsy+8);
    rvtx[5].x = DV(lsx+LASER_SIGHT_WIDTH-9);
    rvtx[5].y = DV(lsy);
    rvtx[6].x = DV(lsx+LASER_SIGHT_WIDTH-1);
    rvtx[6].y = DV(lsy+LASER_SIGHT_WIDTH-1);
    rvtx[7].x = DV(lsx+LASER_SIGHT_WIDTH-9);
    rvtx[7].y = DV(lsy+LASER_SIGHT_WIDTH-1);
    rvtx[8].x = DV(lsx+LASER_SIGHT_WIDTH-1);
    rvtx[8].y = DV(lsy+LASER_SIGHT_WIDTH-9);
    rvtx[9].x = DV(lsx);
    rvtx[9].y = DV(lsy+LASER_SIGHT_WIDTH-1);
    rvtx[10].x = DV(lsx);
    rvtx[10].y = DV(lsy+LASER_SIGHT_WIDTH-9);
    rvtx[11].x = DV(lsx+8);
    rvtx[11].y = DV(lsy+LASER_SIGHT_WIDTH-1);
    rvtx[0].diffuse = rvtx[1].diffuse = rvtx[2].diffuse
      = rvtx[3].diffuse = rvtx[4].diffuse = rvtx[5].diffuse
      = rvtx[6].diffuse = rvtx[7].diffuse = rvtx[8].diffuse
      = rvtx[9].diffuse = rvtx[10].diffuse = rvtx[11].diffuse 
      = lockColor;
    GaDrawPrimitiveRV(rvtx, 12, laserIp, 12);
  }
  lsNum = 0;
}

static DTRI tailIp[12] = {0,1,2, 0,2,1, 1,2,3, 1,3,2};

static void drawTail(Meteo *mt) {
  GaVertex vtx[4];
  GaVect head, bv, plumb;
  DVAL length;
  int i, ti;
  GaMatrix mtr;
  DCOL tailColor;
  int tcPrm;
  int od, td, tmv;
  if ( randN(3) == 0 ) return;
  GaMatIdentity(&mtr);
  GaSetTransform(&mtr);
  GaLightEnable(FALSE);
  head.x = mt->mx; head.y = mt->my; head.z = mt->mz;
  GaVectNormalize(&head);
  head.x <<= 4; head.y <<= 4; head.z <<= 4; 
  vtx[0].x = mt->x + (head.x<<1);
  vtx[0].y = mt->y + (head.y<<1);
  vtx[0].z = mt->z + (head.z<<1);
  bv.z = 0;
  length = GaDiv(DV(32760), (mt->z>>2));
  tailColor = RGBA_MAKE(255, 
			100+randN(155),
			100+randN(155), 0);
  GaSetMaterial(tailColor);
  for ( i=0, ti=0 ; i<=TAIL_NUM ; i++, ti++ ) {
    ti &= (TAIL_NUM-1);
    bv.x = GaSin(D2R(mt->tailDeg[ti]));
    bv.y = GaCos(D2R(mt->tailDeg[ti]));
    GaVectCross(&head, &bv, &plumb);
    if ( i == 0 ) {
      vtx[1].x = mt->x + plumb.x;
      vtx[1].y = mt->y + plumb.y;
      vtx[1].z = mt->z + plumb.z;
    } else {
      od = mt->tailDeg[ti] - mt->tailDeg[i-1];
      if ( od<-180 || (od>0 && od<180) ) {
	vtx[2] = vtx[1];
	vtx[1].x = mt->x + plumb.x;
	vtx[1].y = mt->y + plumb.y;
	vtx[1].z = mt->z + plumb.z;
      } else {
	td = mt->tailDeg[i-1];
	tmv = mt->tailDegMv[i-1];
	mt->tailDeg[i-1] = mt->tailDeg[ti];
	mt->tailDegMv[i-1] = mt->tailDegMv[ti];
	mt->tailDeg[ti] = td;
	mt->tailDegMv[ti] = tmv;
	vtx[2].x = mt->x + plumb.x;
	vtx[2].y = mt->y + plumb.y;
	vtx[2].z = mt->z + plumb.z;
      }
      vtx[3].x = ((vtx[1].x + vtx[2].x)>>1) - GaMul(head.x, length);
      vtx[3].y = ((vtx[1].y + vtx[2].y)>>1) - GaMul(head.y, length);
      vtx[3].z = ((vtx[1].z + vtx[2].z)>>1) - GaMul(head.z, length);
      GaDrawPrimitive(vtx, 4, tailIp, 12);
    }
  }
}

static DCOL meteoColor = RGBA_MAKE(250, 40, 80, 0);

static void drawMeteos() {
  int i, idx;
  GaMatrix mtr;
  GaVertex vtx;
  GaRVertex rvtx;
  Meteo *mt;
  vtx.x = vtx.y = vtx.z = DV(0);
  idx = meteoIdx;
  for ( i=0 ; i<METEO_MAX ; i++ ) {
    idx--; idx &= METEO_MAX-1;
    if ( meteo[idx]->z == NOT_EXIST ) continue;
    GaLightEnable(TRUE);
    GaSetMaterial(meteoColor);
    mt = meteo[idx];
    GaMatIdentity(&mtr);
    GaMatTranslation(&mtr, mt->x, mt->y, mt->z);
    GaMatRotationX(&mtr, D2R(mt->d1));
    GaMatRotationY(&mtr, D2R(mt->d2));
    GaSetTransform(&mtr);
    GaDrawModel(mt->model);
      
    GaTransform(&vtx, &rvtx, 1);
    mt->sx = IDV(rvtx.x);
    mt->sy = IDV(rvtx.y);
    drawTail(mt);
  }
}

static DCOL fragColor[] = {
  RGBA_MAKE(125, 125, 205, 0),
  RGBA_MAKE(125, 50, 200, 0),
  RGBA_MAKE(125, 125, 120, 0),
  RGBA_MAKE(120, 50, 120, 0),
};
static int fragDisp[] = {
  0,1,1,0,1,1,0,1,0,1,0,0,1,0,0,1
};
static DCOL blackColor = RGBA_MAKE(0, 0, 0, 0);

extern GaModel fragModel[];

static void drawFrags() {
  int i;
  GaMatrix mtr;
  Frag *fr;
  int mdlIdx;
  GaLightEnable(FALSE);
  for ( i=0 ; i<FRAG_MAX ; i++ ) {
    if ( frag[i].cnt < 0 ) continue;
    fr = &(frag[i]);
    if ( fr->type==0 && fr->cnt<16 && !fragDisp[fr->cnt] ) continue;
    switch ( fr->type ) {
    case 0:
      GaSetMaterial(fragColor[randN(4)]);
      if ( fr->cnt >= 32 ) {
	mdlIdx = 0;
      } else {
	mdlIdx = 3-(fr->cnt>>3);
      }
      break;
    case 1:
      GaSetMaterial(laserColor[randN(4)]);
      mdlIdx = 1;
      break;
    case 2:
      GaSetMaterial(RGBA_MAKE(200, 96+randN(128), 96+randN(128), 0));
      mdlIdx = 2;
      break;
    }
    if ( gameStatus.status == HIT ) {
      GaSetMaterial(blackColor);
    }
    GaMatIdentity(&mtr);
    GaMatTranslation(&mtr, fr->x, fr->y, fr->z);
    GaMatRotationX(&mtr, D2R(fr->d1));
    GaMatRotationY(&mtr, D2R(fr->d2));
    GaSetTransform(&mtr);
    GaDrawModel(&(fragModel[mdlIdx]));
  }
}

static DTRI whiteIp[] = {0,1,2, 0,2,1, 1,2,3, 1,3,2};

static void drawWhiteout(int cnt) {
  GaRVertex rvtx[12];
  int y;
  DCOL whiteColor;
  y = 160-(cnt<<4); if ( y<0 ) y = 0;
  whiteColor = RGBA_MAKE(255-y, 255-y, 255-y, 0);
  rvtx[0].x = rvtx[2].x = DV(0);
  rvtx[1].x = rvtx[3].x = DV(160);
  rvtx[0].y = rvtx[1].y = DV(y);
  rvtx[2].y = rvtx[3].y = DV(160);
  rvtx[0].diffuse = rvtx[1].diffuse 
    = rvtx[2].diffuse = rvtx[3].diffuse = whiteColor;
  GaDrawPrimitiveRV(rvtx, 4, whiteIp, 12);
}

void draw() {
  switch ( gameStatus.status ) {
  case HIT:
    drawWhiteout(gameStatus.cnt);
    break;
  }
  drawFrags();
  drawMeteos();
  switch ( gameStatus.status ) {
  case GAME_IN_PROGRESS:
    drawLasers();
    drawLaserSights();
    break;
  case TITLE:
    drawTitle(gameStatus.cnt);
    break;
  }
  gameStatus.cnt++;
}
