/*
 * modelmaker.c
 *
 * Copyright (c) 2001- Kenta Cho. All rights reserved.
 */
#include "galib.h"
#include "bslib.h"
#include "meteo.h"

/**
 * Make 3d model datas.
 *
 * @author cs8k-cyu@asahi-net.or.jp
 * @version 0.1
 */

GaModel meteoModel[METEO_MODEL_NUM];
GaModel fragModel[FRAG_MODEL_NUM];

static float meteoVtx[] = {
  0,16,-8, -13,-8,-8, 13,-8,-8, 0,0,16,
};
static int meteoIdx[] = {
  0,2,1, 0,1,3, 1,2,3, 2,0,3,
};

static float fragVtx[] = {
  0,16,0, -13,-8,0, 13,-8,0,
};
static int fragIdx[] = {
  0,1,2, 0,2,1,
};

void makeModel(GaModel* model, 
	       float vtx[], int idx[], int numVtx, int numIdx) {
  int i;
  int vi;
  GaVect vct;
  GaVertex* vertex;
  model->NVertex = numVtx; model->NIndex = numIdx;
  model->Vertexp = MemAlloc(sizeof(GaVertex)*numVtx);
  model->Indexp = MemAlloc(sizeof(unsigned short)*numIdx);
  for ( i=0, vi=0 ; i<numVtx ; i++, vi+=3 ) {
    vertex = &(model->Vertexp[i]);
    vertex->x = DV(vtx[vi]+randN(64)*0.1f);
    vertex->y = DV(vtx[vi+1]+randN(64)*0.1f);
    vertex->z = DV(vtx[vi+2]+randN(64)*0.1f);
    vct.x = vertex->x; vct.y = vertex->y; vct.z = vertex->z;
    GaVectNormalize(&vct);
    vertex->nx = vct.x; vertex->ny = vct.y; vertex->nz = vct.z;
    vertex->tu = vertex->tv = DV(0);
  }
  for ( i=0 ; i<numIdx ; i++ ) {
    model->Indexp[i] = idx[i];
  }
}

void makeModels() {
  int i, j;
  for ( i=0 ; i<METEO_MODEL_NUM ; i++ ) {
    makeModel(&(meteoModel[i]), meteoVtx, meteoIdx, 4, 12);
  }
  for ( i=0 ; i<FRAG_MODEL_NUM ; i++ ) {
    makeModel(&(fragModel[i]), fragVtx, fragIdx, 3, 6);
    for ( j=0 ; j<9 ; j++ ) {
      fragVtx[j] *= 0.8;
    }
  }
}
